# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from swaggerit.utils import get_model_schema
from sqlalchemy.ext.declarative import AbstractConcreteBase, declared_attr
from base64 import b64decode
import sqlalchemy as sa
import binascii
import re


class GrantsModelBase(AbstractConcreteBase):
    __tablename__ = 'grants'
    __use_redis__ = False

    @declared_attr
    def uri_id(cls):
        return sa.Column(sa.ForeignKey('uris.id'), primary_key=True)

    @declared_attr
    def method_id(cls):
        return sa.Column(sa.ForeignKey('methods.id'), primary_key=True)

    @declared_attr
    def uri(cls):
        return sa.orm.relationship('URIsModel')

    @declared_attr
    def method(cls):
        return sa.orm.relationship('MethodsModel')


class URIsModelBase(AbstractConcreteBase):
    __tablename__ = 'uris'
    __use_redis__ = False

    id = sa.Column(sa.Integer, primary_key=True)
    uri = sa.Column(sa.String(255), unique=True, nullable=False)


class MethodsModelBase(AbstractConcreteBase):
    __tablename__ = 'methods'
    __use_redis__ = False

    id = sa.Column(sa.Integer, primary_key=True)
    method = sa.Column(sa.String(10), unique=True, nullable=False)


class UsersModelBase(AbstractConcreteBase):
    __tablename__ = 'users'
    __schema__ = get_model_schema(__file__)

    id = sa.Column(sa.String(255), primary_key=True)
    name = sa.Column(sa.String(255), unique=True, nullable=False)
    email = sa.Column(sa.String(255), unique=True, nullable=False)
    password = sa.Column(sa.String(255), nullable=False)
    admin = sa.Column(sa.Boolean, default=False)

    @declared_attr
    def grants(cls):
        grants_primaryjoin = 'UsersModel.id == users_grants.c.user_id'
        grants_secondaryjoin = 'and_('\
            'GrantsModel.uri_id == users_grants.c.grant_uri_id, '\
            'GrantsModel.method_id == users_grants.c.grant_method_id)'

        return sa.orm.relationship(
            'GrantsModel', uselist=True, secondary='users_grants',
            primaryjoin=grants_primaryjoin, secondaryjoin=grants_secondaryjoin)

    @declared_attr
    def stores(cls):
        return sa.orm.relationship('StoresModel', uselist=True, secondary='users_stores')

    @classmethod
    async def authorize(cls, session, authorization, url, method):
        try:
            authorization = b64decode(authorization).decode()
        except binascii.Error:
            return None
        except UnicodeDecodeError:
            return None

        if not ':' in authorization:
            return None

        user = await cls.get(session, {'id': authorization})
        user = user[0] if user else user
        if user and user.get('admin'):
            session.user = user
            return True

        elif user:
            if method == 'OPTIONS':
                return True

            for grant in user['grants']:
                grant_uri = grant['uri']['uri']
                if (grant_uri == url or re.match(grant_uri, url)) \
                        and grant['method']['method'].lower() == method.lower():
                    session.user = user
                    return True

            return False

    @classmethod
    async def insert(cls, session, objs, commit=True, todict=True):
        objs = cls._to_list(objs)
        await cls._set_objs_ids_and_grant(objs, session)
        return await type(cls).insert(cls, session, objs, commit, todict)

    @classmethod
    async def _set_objs_ids_and_grant(cls, objs, session):
        objs = cls._to_list(objs)

        patch_method = await cls.get_model('methods').get(session, ids={'method': 'patch'}, todict=False)
        if not patch_method:
            patch_method = await cls.get_model('methods').insert(session, [{'method': 'patch'}], todict=False)
        patch_method = patch_method[0]

        get_method = await cls.get_model('methods').get(session, ids={'method': 'get'}, todict=False)
        if not get_method:
            get_method = await cls.get_model('methods').insert(session, [{'method': 'get'}], todict=False)
        get_method = get_method[0]

        for obj in objs:
            new_grants = []
            user_uri = '/users/{}'.format(obj['email'])

            uri = await cls.get_model('uris').get(session, ids={'uri': user_uri}, todict=False)
            if not uri:
                uri = await cls.get_model('uris').insert(session, [{'uri': user_uri}], todict=False)
            uri = uri[0]

            grant = await cls.get_model('grants').get(session, {'uri_id': uri.id, 'method_id': patch_method.id}, todict=False)
            if grant:
                grant = grant[0].todict()
            else:
                grant = {'uri_id': uri.id, 'method_id': patch_method.id, '_operation': 'insert'}
            new_grants.append(grant)

            grant = await cls.get_model('grants').get(session, {'uri_id': uri.id, 'method_id': get_method.id}, todict=False)
            if grant:
                grant = grant[0].todict()
            else:
                grant = {'uri_id': uri.id, 'method_id': get_method.id, '_operation': 'insert'}
            new_grants.append(grant)

            obj['id'] = '{}:{}'.format(obj['email'], obj['password'])
            grants = obj.get('grants', [])
            grants.extend(new_grants)
            obj['grants'] = grants

    @classmethod
    async def update(cls, session, objs, commit=True, todict=True, ids=None, ids_keys=None):
        if not ids:
            ids = []
            objs = cls._to_list(objs)
            for obj in objs:
                id_ = obj.get('id')
                email = obj.get('email')
                if id_ is not None:
                    ids.append({'id': id_})
                    ids_keys = ('id',)
                elif email is not None:
                    ids.append({'email': email})
                    ids_keys = ('email',)

        insts = await type(cls).update(cls, session, objs, commit=False,
                            todict=False, ids=ids, ids_keys=ids_keys)
        cls._set_insts_ids(insts)

        if commit:
            await session.commit()
        return cls._build_todict_list(insts) if todict else insts

    @classmethod
    def _set_insts_ids(cls, insts):
        insts = cls._to_list(insts)
        for inst in insts:
            inst.id = '{}:{}'.format(inst.email, inst.password)


def build_users_grants_table(metadata, **kwargs):
    return sa.Table(
        'users_grants', metadata,
        sa.Column('user_id', sa.String(255), sa.ForeignKey('users.id', ondelete='CASCADE', onupdate='CASCADE'), primary_key=True),
        sa.Column('grant_uri_id', sa.Integer, sa.ForeignKey('grants.uri_id', ondelete='CASCADE', onupdate='CASCADE'), primary_key=True),
        sa.Column('grant_method_id', sa.Integer, sa.ForeignKey('grants.method_id', ondelete='CASCADE', onupdate='CASCADE'), primary_key=True),
        **kwargs)


def build_users_stores_table(metadata, **kwargs):
    return sa.Table(
        'users_stores', metadata,
        sa.Column('user_id', sa.String(255), sa.ForeignKey('users.id', ondelete='CASCADE', onupdate='CASCADE'), primary_key=True),
        sa.Column('store_id', sa.Integer, sa.ForeignKey('stores.id', ondelete='CASCADE', onupdate='CASCADE'), primary_key=True),
        **kwargs)



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.engines.cores.base import AbstractDataImporter, EngineCore
from myreco.engines.cores.utils import build_engine_data_path
from unittest import mock
from jsonschema import ValidationError
from os import makedirs
import os.path
import gzip
import ujson
import asyncio


class DataImporterTest(AbstractDataImporter):

    def get_data(self, items_indices_map, session):
        asyncio.run_coroutine_threadsafe(
            asyncio.sleep(0.5),
            session.loop
        )
        data_path = build_engine_data_path(self._engine)
        if not os.path.isdir(data_path):
            makedirs(data_path)

        data = [{'item_key': '2|test2', 'value': 1},
                {'item_key': '1|test1', 'value': 3},
                {'item_key': '3|test3', 'value': 2}]
        data = map(ujson.dumps, data)
        data = '\n'.join(data)

        filename_prefix = 'top_seller'
        file_ = gzip.open(os.path.join(data_path, filename_prefix) + '-000000001.gz', 'wt')
        file_.write(data)
        file_.close()
        return {'lines_count': 3}


def CoroMock():
    coro = mock.MagicMock(name="CoroutineResult")
    corofunc = mock.MagicMock(name="CoroutineFunction", side_effect=asyncio.coroutine(coro))
    corofunc.coro = coro
    return corofunc


class EngineCoreTest(EngineCore):
    __configuration_schema__ = {
        "type": "object",
        "required": ["item_id_name", "aggregators_ids_name"],
        "properties": {
            "item_id_name": {"type": "string"},
            "aggregators_ids_name": {"type": "string"}
        }
    }

    def get_variables(self):
        item_id_name = self.engine['configuration']['item_id_name']
        aggregators_ids_name = self.engine['configuration']['aggregators_ids_name']
        item_type_schema_props = self.engine['item_type']['schema']['properties']
        return [{
            'name': item_id_name,
            'schema': item_type_schema_props[item_id_name]
        },{
            'name': aggregators_ids_name,
            'schema': item_type_schema_props[aggregators_ids_name]
        }]

    def _validate_config(self, engine):
        item_id_name = engine['configuration']['item_id_name']
        aggregators_ids_name = engine['configuration']['aggregators_ids_name']
        item_type_schema_props = engine['item_type']['schema']['properties']
        message = "Configuration key '{}' not in item_type schema"

        if item_id_name not in item_type_schema_props:
            raise ValidationError(message.format('item_id_name'),
                instance=engine['configuration'], schema=item_type_schema_props)

        elif aggregators_ids_name not in item_type_schema_props:
            raise ValidationError(message.format('aggregators_ids_name'),
                instance=engine['configuration'], schema=item_type_schema_props)

    async def _build_rec_vector(self):
        pass

    get_recommendations = CoroMock()



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from swaggerit.utils import build_validator, get_module_path, set_logger
from swaggerit.json_builder import JsonBuilder
from myreco.engines.cores.items_indices_map import ItemsIndicesMap
from myreco.engines.cores.utils import build_engine_data_path, build_engine_key_prefix
from jsonschema import Draft4Validator
from abc import ABCMeta, abstractmethod
from bottleneck import argpartition
from glob import glob
from aiofiles import gzip_open
import ujson
import os.path


class EngineError(Exception):
    pass


class EngineCoreMeta(ABCMeta):

    def __init__(cls, name, bases_classes, attributes):
        if name != 'EngineCore':
            schema = cls.__configuration_schema__
            Draft4Validator.check_schema(schema)
            cls.__config_validator__ = build_validator(schema, get_module_path(cls))


class EngineCore(metaclass=EngineCoreMeta):

    def __init__(self, engine=None, items_model=None):
        self.engine = engine
        self.items_model = items_model
        set_logger(self, build_engine_key_prefix(self.engine))

    def get_variables(self):
        return []

    def validate_config(self):
        self.__config_validator__.validate(self.engine['configuration'])
        self._validate_config(self.engine)

    def _validate_config(self, engine):
        pass

    async def get_recommendations(self, session, filters, max_recos, show_details, **variables):
        rec_vector = await self._build_rec_vector(session, **variables)

        if rec_vector is not None:
            [await filter_.filter(session, rec_vector, ids) for filter_, ids in filters.items()]
            return await self._build_rec_list(session, rec_vector, max_recos, show_details)

        return []

    @abstractmethod
    async def _build_rec_vector(self, session, **variables):
        pass

    async def _build_rec_list(self, session, rec_vector, max_recos, show_details):
        items_indices_map = ItemsIndicesMap(self.items_model)
        best_indices = self._get_best_indices(rec_vector, max_recos)
        best_items_keys = await items_indices_map.get_items(best_indices, session)

        if show_details and best_items_keys:
            return [ujson.loads(item) for item in await session.redis_bind.hmget(
                            self.items_model.__key__, *best_items_keys) if item is not None]

        else:
            items_ids = []
            for key in best_items_keys:
                item = {}
                self.items_model.set_instance_ids(item, key)
                self._set_item_values(item)
                items_ids.append(item)

            return items_ids

    def _get_best_indices(self, rec_vector, max_recos):
        if max_recos > rec_vector.size:
            max_recos = rec_vector.size

        best_indices = argpartition(-rec_vector, max_recos-1)[:max_recos]
        best_values = rec_vector[best_indices]
        return [int(i) for i, v in
            sorted(zip(best_indices, best_values), key=lambda x: x[1], reverse=True) if v > 0.0]

    def _set_item_values(self, item):
        for k in item:
            schema = self.engine['item_type']['schema']['properties'].get(k)
            if schema is None:
                raise EngineError('Invalid Item {}'.format(item))

            item[k] = JsonBuilder.build(item[k], schema)

    def export_objects(self, session):
        pass

    async def _build_csv_readers(self, pattern):
        path = build_engine_data_path(self.engine)
        readers = []
        for filename in glob(os.path.join(path, '{}*.gz'.format(pattern))):
            file_ = await gzip_open(filename, 'rt')
            readers.append(file_)
        return readers

    async def _get_items_indices_map_dict(self, items_indices_map, session):
        items_indices_map_dict = await items_indices_map.get_all(session)

        if not items_indices_map_dict.values():
            raise EngineError(
                "The Indices Map for '{}' is empty. Please update these items"
                .format(self.engine['item_type']['name']))

        return items_indices_map_dict


class AbstractDataImporter(metaclass=ABCMeta):

    def __init__(self, engine):
        self._engine = engine

    @abstractmethod
    def get_data(cls, items_indices_map, session):
        pass


class RedisObjectBase(object):

    def __init__(self, engine_core):
        self._engine_core = engine_core
        self._set_redis_key()
        set_logger(self, self._redis_key)

    def _set_redis_key(self):
        self._redis_key = build_engine_key_prefix(self._engine_core.engine)



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.engines.cores.items_indices_map import ItemsIndicesMap
from zlib import decompress, compress
from collections import defaultdict
import numpy as np


class FilterBaseBy(object):
    dtype = np.bool

    def __init__(self, items_model, name, is_inclusive=True, id_names=None, skip_values=None):
        self.key = items_model.__key__ + '_' + name + '_filter'
        self.items_model = items_model
        self.name = name
        self.is_inclusive = is_inclusive
        self.id_names = id_names
        self.skip_values = set(skip_values) if skip_values is not None else skip_values

    def _unpack_filter(self, filter_, new_size=None):
        filter_ = np.fromstring(filter_, dtype=type(self).dtype)

        if new_size is not None:
            self._resize_vector(filter_, new_size)

        return filter_

    def _resize_vector(self, vector, new_size):
        if vector.size != new_size:
            vector.resize(new_size, refcheck=False)

    def _filter(self, filter_, rec_vector):
        if not self.is_inclusive:
            filter_ = np.invert(filter_)

        rec_vector *= filter_

    def _pack_filter(self, filter_):
        return filter_.tobytes()

    def _build_empty_array(self, size):
        return np.zeros(size, dtype=np.bool)

    def _build_output_ids(self, item):
        ids = [item.get(id_name) for id_name in self.items_model.__id_names__]
        return ' | '.join([str(i) for i in ids])

    def _list_cast(self, obj):
        return obj if isinstance(obj, list) or isinstance(obj, tuple) else (obj,)

    def _not_skip_value(self, value):
        return (self.skip_values is None or value not in self.skip_values)

    async def _get_items(self, session, items_ids):
        keys = [self.items_model.get_instance_key(item_id) for item_id in items_ids]
        return await self.items_model.get(session, keys)


class BooleanFilterBy(FilterBaseBy):

    async def update(self, session, items):
        filter_ = self._build_empty_array(len(items))

        for item in items:
            value = item.get(self.name)
            if value is not None and self._not_skip_value(value):
                filter_[item['index']] = value

        await session.redis_bind.set(self.key, self._pack_filter(filter_))
        return {'true_values': np.nonzero(filter_)[0].size}

    async def filter(self, session, rec_vector, *args, **kwargs):
        filter_ = await session.redis_bind.get(self.key)
        if filter_ is not None:
            filter_ = self._unpack_filter(filter_, rec_vector.size)
            self._filter(filter_, rec_vector)


class MultipleFilterBy(FilterBaseBy):

    async def filter(self, session, rec_vector, ids):
        ids = self._list_cast(ids)

        if ids:
            filters = await session.redis_bind.hmget(self.key, *ids)
            filters = [self._unpack_filter(filter_, rec_vector.size)
                        for filter_ in filters if filter_ is not None]
            final_filter = np.ones(rec_vector.size, dtype=np.bool)

            for filter_ in filters:
                final_filter *= filter_

            self._filter(final_filter, rec_vector)

    async def update(self, session, items):
        filter_map = defaultdict(list)
        set_data = dict()
        size = len(items)

        [self._update_filter(filter_map, item) for item in items]
        for filter_id, items_indices in filter_map.items():
            filter_ = self._build_filter_array(items_indices, size)
            set_data[filter_id] = self._pack_filter(filter_)

        if set_data:
            await session.redis_bind.hmset_dict(self.key, set_data)

        return {'filters_quantity': len(set_data)}

    def _update_filter(self, filter_map, value, item):
        if value is not None and self._not_skip_value(value):
            filter_map[value].append(item['index'])

    def _build_filter_array(self, items_indices, size):
        filter_ = self._build_empty_array(size)
        filter_[np.array(items_indices, dtype=np.int32)] = True
        return filter_


class SimpleFilterBy(MultipleFilterBy):

    def _update_filter(self, filter_map, item):
        MultipleFilterBy._update_filter(self, filter_map, item.get(self.name), item)


class ObjectFilterBy(MultipleFilterBy):

    def _update_filter(self, filter_map, item):
        value = self._get_id_from_property(item)
        MultipleFilterBy._update_filter(self, filter_map, value, item)

    def _get_id_from_property(self, item):
        property_obj = item.get(self.name)
        if property_obj is not None:
            ids = [property_obj[id_name] for id_name in self.id_names]
            return repr(tuple([id_ for _, id_ in sorted(zip(self.id_names, ids), key=lambda x: x[0])]))

    async def filter(self, session, rec_vector, properties):
        properties = self._list_cast(properties)
        ids = [self._get_id_from_property({self.name: prop}) for prop in properties]
        await MultipleFilterBy.filter(self, session, rec_vector, ids)


class ArrayFilterBy(SimpleFilterBy):

    def _update_filter(self, filter_map, item):
        property_list = item.get(self.name, [])
        for value in property_list:
            MultipleFilterBy._update_filter(self, filter_map, value, item)


class SimpleFilterOf(SimpleFilterBy):

    async def filter(self, session, rec_vector, items_ids):
        items = await self._get_items(session, items_ids)
        filter_ids = [item.get(self.name) for item in items]
        await SimpleFilterBy.filter(self, session, rec_vector, filter_ids)


class ObjectFilterOf(ObjectFilterBy):

    async def filter(self, session, rec_vector, items_ids):
        items = await self._get_items(session, items_ids)
        filter_ids = [item.get(self.name) for item in items]
        await ObjectFilterBy.filter(self, session, rec_vector, filter_ids)


class ArrayFilterOf(ArrayFilterBy):

    async def filter(self, session, rec_vector, items_ids):
        items = await self._get_items(session, items_ids)
        filter_ids = []
        [filter_ids.extend(item[self.name]) for item in items]
        await ArrayFilterBy.filter(self, session, rec_vector, filter_ids)


class IndexFilterOf(FilterBaseBy):

    async def update(self, *args, **kwargs):
        return 'OK'

    async def filter(self, session, rec_vector, items_ids):
        items_indices_map = ItemsIndicesMap(self.items_model)
        indices = await items_indices_map.get_indices(items_ids, session)
        if indices:
            indices = np.array(indices, dtype=np.int32)
            self._filter_by_indices(rec_vector, indices)

    def _filter_by_indices(self, rec_vector, indices):
        if not self.is_inclusive:
            rec_vector[indices] = 0
        else:
            rec_vector[:] = 0
            rec_vector[indices] = 1


class IndexFilterByPropertyOf(SimpleFilterOf, IndexFilterOf):
    dtype = np.int32

    def _build_filter_array(self, items_indices, size):
        return np.array(items_indices, dtype=np.int32)

    async def filter(self, session, rec_vector, items_ids):
        items = await self._get_items(session, items_ids)
        filter_ids = [item[self.name] for item in items]
        if filter_ids:
            filters = await session.redis_bind.hmget(self.key, *filter_ids)
            filters = [self._unpack_filter(filter_) for filter_ in filters if filter_ is not None]

            if filters:
                indices = np.concatenate(filters)
                self._filter_by_indices(rec_vector, indices)



# Copyright 2016 Diogo Dutra

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from version import VERSION
from setuptools import setup, find_packages


long_description = ''
with open('README.md') as readme:
    long_description = readme.read()


dependency_links = []
install_requires = []
with open('requirements.txt') as requirements:
    lines = requirements.readlines()
    install_requires = lines[2:]
    dependency_links = lines[:2]
    del lines


tests_require = []
with open('requirements-dev.txt') as requirements_dev:
    tests_require = requirements_dev.readlines()


setup(
    name='myreco',
    packages=find_packages('.'),
    include_package_data=True,
    version=VERSION,
    description='A Recommendations Framework',
    long_description=long_description,
    author='Diogo Dutra',
    author_email='dutradda@gmail.com',
    url='https://github.com/dutradda/myreco',
    download_url='http://github.com/dutradda/myreco/archive/master.tar.gz',
    license='MIT',
    keywords='recommendations neighborhood visualsimilarity visual similarity topseller top seller'\
    		' swagger openapi falconframework falcon framework',
    setup_requires=[
        'pytest-runner==2.9',
        'setuptools==28.3.0'
    ],
    tests_require=tests_require,
    install_requires=install_requires,
    dependency_links=dependency_links,
    classifiers=[
    	'License :: OSI Approved :: MIT License',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)
